// sample(List,S) : Samples S from list L
// e.g. sample([1,2,3],S) - S will contain one number from [1,2,3] with uniform probability
sample(List,S) :- length(List,L) & sample(List,L,S).

sample([],0,undefined).
sample([X],1,X).
sample([X|Xs],Length,X) :- math.random < 1/Length.
sample([_|Xs],Length,R) :- sample(Xs,Length-1,R).


// length(List,L) : Unifies L with length of List
length([],0).
length([_|Xs],L) :- length(Xs,L1) & L=L1+1.

// filter(List,Not,Filtered) : Filtered contains all elements of List not equal to Not
// e.g. filter([1,2,3],2,X) - X=[1,3]
filterNot([],_,[]).
filterNot([X|Xs],X,Filtered) :- filterNot(Xs,X,Filtered).
filterNot([X|Xs],Not,[X|Filtered]) :- filterNot(Xs,Not,Filtered).

// head(List,Result) - Result contains first element of List or false if the list is empty
head([],false).
head(List,List).

// opposite(Direction,Opposite) - Opposite contains opposite direction to Direction
opposite(left,right).
opposite(right,left).
opposite(up,down).
opposite(down,up).

// next(Xc,Yc,Direction,Xn,Yn) - Let (Xc,Yc) be current position. (Xn,Yn) is the next position
//                               when applying action in the direction specified by Direction
next(Xc,Yc,left,Xc-1,Yc) :- valid(Xc-1,Yc).
next(Xc,Yc,right,Xc+1,Yc) :- valid(Xc+1,Yc).
next(Xc,Yc,up,Xc,Yc-1) :- valid(Xc,Yc-1).
next(Xc,Yc,down,Xc,Yc+1) :- valid(Xc,Yc+1).

// valid(X,Y) - holds if X,Y lies within the map boundary
valid(X,Y) :- gsize(_,W,H) & X >= 0 & X < W & Y >= 0 & Y < H.


// in_4N(X1,Y1,X2,Y2) - holds if (X1,Y1) is in 4-neighbourhood of (X2,Y2)
in_4N(X,Y,X+1,Y) :- valid(X,Y) & valid(X+1,Y).
in_4N(X,Y,X-1,Y) :- valid(X,Y) & valid(X-1,Y).
in_4N(X,Y,X,Y+1) :- valid(X,Y) & valid(X,Y+1).
in_4N(X,Y,X,Y-1) :- valid(X,Y) & valid(X,Y-1).

// in_8N(X1,Y1,X2,Y2) - holds if (X1,Y1) is in 8-neighbourhood of (X2,Y2)
in_8N(X,Y,X2,Y2) :- in_4N(X,Y,X2,Y2).
in_8N(X,Y,X+1,Y+1) :- valid(X,Y) & valid(X+1,Y+1).
in_8N(X,Y,X-1,Y+1) :- valid(X,Y) & valid(X-1,Y+1).
in_8N(X,Y,X+1,Y-1) :- valid(X,Y) & valid(X+1,Y-1).
in_8N(X,Y,X-1,Y-1) :- valid(X,Y) & valid(X-1,Y-1).

// dist(X1,Y1,X2,Y2,D) - D unifies with Manhattan distance between (X1,Y1) and (X2,Y2)
dist(X1,Y1,X2,Y2,D) :- D=(math.abs(X1-X2) + math.abs(Y1-Y2)).

// miners(M) - M unifies with the list of miners' names
miners([miner1,miner2,miner3,miner4]).

!run.

+?random_move(left)  : math.random < 0.250.
+?random_move(right) : math.random < 0.333.
+?random_move(up)    : math.random < 0.500.
+?random_move(down).


+!run <- !step ; !run.
+!step <- ?random_move(D) ; do(D).
